/*  $Id: AStaticTranslator.java 1335 2012-08-25 13:27:49Z sdevaux $
 *
 * Tabs = 4 chars
 * copyright (C) 2002-2012 Sebastien Devaux
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 *
 *              Sebastien Devaux <sebastien.devaux@laposte.net>
 */
package net.eduvax.util;

import java.util.Hashtable;
import java.util.StringTokenizer;
/** 
 * Convert utilities based on static table.
 */
public abstract class AStaticTranslator<T,U> implements Translator<T,U> {
	/** Convertion map */
	private Hashtable<T,U> _table;
	/** 
	 * Default value.
	 * Fallback value to return when entry is not found in table.
	 */
	private U _nullValue;
	/** 
	 * Create new static translator.
	 * @param nullVal default value to return in response of unknown
	 * entry.
	 */
	public AStaticTranslator(U nullVal)
	{
	    _nullValue=nullVal;
	    _table=new Hashtable<T,U>();
	}
	/** 
	 * String to key convertion.
	 * @param str string to convert.
	 * @return string converted to key type.
	 */ 
	protected abstract T sourceTypeFromString(String str) 
						throws Exception;
	/** 
	 * String to value convertion.
	 * @param str string to convert.
	 * @return string converted to value type. 
	 */
	protected abstract U targetTypeFromString(String str)
						throws Exception;
	/** 
	 * Fill table using a string definition.
	 * @param tblText Table textual description. Each table entry is
	 * written as : source=target, separated with ';' char. Tab and
	 * carriage return are ignored. Example
	 * <pre>1=One; 2=Two;
	 * 3=Three; 4=Four;
	 * 5=Five</pre>
	 * @param eh error handler to use in case of exception while
	 * converting elements to U and T type.
	 */ 
	public void fillTable(String tblText,ErrorHandler eh) {
		ErrorHandler errHandler=eh!=null?eh:
					App.get().getErrHandler();
		StringTokenizer st=new StringTokenizer(tblText,";");
		while (st.hasMoreTokens()) {
			String rule=st.nextToken();
			try {
				int equPos=rule.indexOf('=');
				if (equPos>0) {
					T key=sourceTypeFromString(
						rule.substring(0,equPos).trim());
					U val=targetTypeFromString(
						rule.substring(equPos+1).trim());
					_table.put(key,val);
				}
			}
			catch (Exception ex) {
				Error err=new Error(""+getClass()+".fill",
                                        Error.ELevel.WARN,ex);
                err.addArg(rule);
				errHandler.handle(err);
			}
	   	}
	}
	// --- Translator interface implementation ---
	public U translate(T toTranslate) {
		U res=_table.get(toTranslate);
		if (res==null) {
			res=_nullValue;
		}
        return res;
	}
}
