/*  $Id: ScriptRunner.java 1335 2012-08-25 13:27:49Z sdevaux $
 *
 * Tabs = 4 chars
 * copyright (C) 2002-2012 Sebastien Devaux
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 *
 *              Sebastien Devaux <sebastien.devaux@laposte.net>
 */
package net.eduvax.util;

import java.io.FileReader;
import javax.script.Bindings;
import javax.script.ScriptContext;
import javax.script.ScriptEngineManager;
import javax.script.ScriptEngine;

/** Gestionnaire d'exécution du scripts./
 * <p>Pour le moment, seul le langage python est pris en charge via
 * l'interpréteur IronPython. Cependant ce dernier est basé sur le moteur
 * Microsoft.Scripting qui est conçu pour accepter plusieurs langage (ruby
 * est déjà disponible avec IronRuby).</p>
 * <p>La mise en oeuvre proposée ici est un peu particulière pour que les
 * interpréteurs soient intégré au logiciel utilisateur et éviter toute 
 * installation fastidieuse d'un produit tiers. En outre, par ce choix on
 * permet une installation identique du logiciel aussi bien sur .Net
 * Windows, que sur Mono Linux (on évite l'enregistrement du composant via
 * gacutil).</p>
 * <p>Par contre, on court circuite les services automatique de séléction
 * du bon interpréteur en fonction de la nature script soumis. Il est donc
 * nécessaire de gérer l'instanciation explicite de chaque moteur pour 
 * chaque langage de script. C'est pourquoi le final staticructeur est restreint
 * et la création accessible par la méthode Factory getScriptRunner. Cela
 * facilitera l'ajout de la prise en charge d'autres langages par héritage
 * de la présente classe.</p>
 * <p>Des objets peuvent être transmis au script grâce à la méthode
 * ExportVar. Cet export doit être fait avant le début de l'execution. Le
 * singleton CApp.App est exporté sous le nom _app.</p>
 */
public class ScriptRunner implements Runnable
{
	/** Moteur de l'interpréteur. */
	private ScriptEngine _engine;
	/** Arguments à transmettre au script */
	private String _args;
	/** Chemin de script à executer. */
	private String _scriptPath;
	/** Envuronnement d'execution du script. */
	private Bindings _bindings;
	/** Gestionnaire d'erreur */
	private ErrorHandler _errHandler;
	/** 
	 * Création d'un nouveau gestionnaire d'execution de script.
	 * @param scriptPath Chemin du script à executer
	 */
	protected ScriptRunner(String scriptPath,String engineName)
	{
		_scriptPath=scriptPath;
		ScriptEngineManager manager=new ScriptEngineManager();
		_errHandler=App.get().getErrHandler();
	
		_engine=manager.getEngineByName(engineName);
		_bindings=_engine.createBindings();
		exportVar("_app",App.get());
	}
	/** Ajouter un argument à transmettre au script */
	public void AddArg(String arg)
	{
		_args=_args+" "+arg;
	}
	/** 
	 * Definir le gestionnaire d'erreur. 
	 * <p>Toute erreur d'execution non gérée dans le script lui-même 
	 * "remonte" dans ce gestionnaire sous forme d'exception. Lors
	 * de l'exécution, toutes les exceptions sont capturée pour être
	 * encapsulée dans un CError qui sera confié au gestionnaire
	 * d'erreur.</p>
	 * @param handler Gestionnaire d'erreur chargé de traiter
	 * les erreurs d'execution du script.
	 */
	public void setErrHandler(ErrorHandler handler)
	{
		_errHandler=handler;
	}
	/** 
	 * Exporter un objet dans l'environnement du script .
	 * @param name Nom de la variable qui sera allouée pour 
	 * l'objet.
	 * @param obj Object à exporter
	 */
	public void exportVar(String name,Object obj) 
	{
		_bindings.put(name,obj);
	}
	/** 
	 * Démarrer l'exécution du script 
	 * <p>Mise en oeuvre IRunnable.</p>
	 */
	public void run() {
		try 
		{
			_engine.setBindings(_bindings,ScriptContext.ENGINE_SCOPE);
			FileReader reader=new FileReader(_scriptPath);
			_engine.eval(reader);
		}
		catch (Exception ex)
		{
			Error err=new Error(getClass().getName(),Error.ELevel.ERR,ex);
			err.addArg(_scriptPath);
			_errHandler.handle(err);
		}
	}
	/** 
	 * Demander la création d'un nouveau gestionnaire d'execution
	 * de script. 
	 * @param script file path
	 * @return a new script runner for this file script.
	 */
	static public ScriptRunner getScriptRunner(String scriptPath)
	{
		String engineName="javascript";
		if (scriptPath.toLowerCase().endsWith(".py")) {
			engineName="python";
		}
		return new ScriptRunner(scriptPath,engineName);
	}
}
